# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cobra_db', 'cobra_db.scripts']

package_data = \
{'': ['*']}

install_requires = \
['deid>=0.2.36,<0.3.0',
 'dnspython>=2.2.1,<3.0.0',
 'numpy>=1.23.2,<2.0.0',
 'pyaml-env>=1.1.5,<2.0.0',
 'pycryptodome>=3.15.0,<4.0.0',
 'pydicom>=2.3.0,<3.0.0',
 'pymongo[srv]>=4.2.0,<5.0.0',
 'tqdm>=4.64.1,<5.0.0']

entry_points = \
{'console_scripts': ['cobra_group = cobra_db.scripts.stage_2:cli',
                     'cobra_ingest = '
                     'cobra_db.scripts.stage_1_ingest_images:cli',
                     'cobra_pseudonymize = '
                     'cobra_db.scripts.pseudonymize_image_metadata:main']}

setup_kwargs = {
    'name': 'cobra-db',
    'version': '0.0.4',
    'description': 'COnsolidated BReast cancer Analysis DataBase',
    'long_description': '# <img src="static/img/cobra_db.png" alt="" height="30"/> cobra_db\n\n\n\n<center>\n\n[![PyPI version](https://badge.fury.io/py/cobra_db.svg)](https://badge.fury.io/py/cobra_db)\n[![Documentation Status](https://readthedocs.org/projects/cobra-db/badge/?version=latest)](https://cobra-db.readthedocs.io/en/latest/?badge=latest)\n[![codecov](https://codecov.io/gh/mammoai/cobra-db/branch/main/graph/badge.svg?token=ASQPS89408)](https://codecov.io/gh/mammoai/cobra-db)\n\n<img src="static/img/cobra_db.png" alt="" height="200"/>\n\n**Co**nsolidated **Br**east Cancer **A**nalysis **D**ata**B**ase\n</center>\n\n\n\n\n## What is ```cobra_db```?\n_cobra_db_ is a python package that allows you to extract DICOM metadata and store it in a MongoDB database. Allowing you to index, transform, and export your medical imaging metadata.\n\nWith cobra_db, you will have more visibility of your data enabling you to get more value from your medical imaging studies.\n\nOnce the metadata is in the database, you can import other text-based information (csv or json) into a custom collection and then run queries. This allows you to mix and match data extracted from different sources in different formats.\n\nFor example, let\'s say you have 1 million mammography DICOM files and you would like to obtain the path of the files that belong to women scanned at an age of between 40 and 50 years old.\n\nIf you had cobra_db, you could run the following query in just a few seconds directly in the mongo shell.\n\n```javascript\ndb.ImageMetadata.find(\n  // filter the data\n  {patient_age:{$gt:40, $lte:50}},\n  // project it into a flat structure\n  {\n    patient_id: "$dicom_tags.PatientID.Value"\n    drive_name: "$file_source.drive_name",\n    rel_path:"$file_source.rel_path",\n  })\n```\nThis would return the patient id, the drive name and the relative path (to the drive) for all the files that match the selection criteria.\n\n## Installation\nIf you already have a working instance of the database, you only need to install the python package.\n\n```bash\n$ pip install cobra_db\n```\n\nIf you would like to create a database from scratch, go ahead and follow the [tutorial](https://cobra-db.readthedocs.io/en/latest/tutorial.html).\n\n## Usage\n\nIf you have an `ImageMetadata` instance id that you would like to access from python.\n\n```python\nfrom cobra_db import Connector, ImageMetadataDao\n\n# the _id of the ImageMetadata instance that you want to access\nim_id = \'62de8e38dc2414586e4ddb25\'\n\n# prompt user for password\nconnector = Connector.get_pass(\n  host=\'my_host.server.com\',\n  port=27017,\n  db_name=\'cobra_db\',\n  username=\'my_user\'\n)\n# connect to the ImageMetadata collection\nim_dao = ImageMetadataDao(connector)\nim = im_dao.get_by_id(im_id)\nprint(im.date.file_source.rel_path)\n\n# this will return\n... rel/path/to/my_file.dcm\n```\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a [Code of Conduct](https://cobra-db.readthedocs.io/en/latest/conduct.html). By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`cobra_db` was created by Fernando Cossio, Apostolia Tsirikoglou, Annika Gregoorian, Haiko Schurz, and Fredrik Strand. It is licensed under the terms of the Apache License 2.0 license.\n\n## Aknowledgements\n\nThis project has been funded by research grants Regional Cancer Centers in Collaboration 21/00060, and Vinnova 2021-0261.\n',
    'author': 'Fernando Cossio',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import os
import sys
from argparse import ArgumentParser, Namespace
from multiprocessing import Pool
from typing import List, Union

import pydicom
from bson import ObjectId
from pyaml_env import parse_config
from pydicom.errors import BytesLengthException
from pymongo.errors import DuplicateKeyError
from tqdm import tqdm

from cobra_db.dataset_mod import DatasetMod
from cobra_db.deid import Deider, base_recipe_path, mr_recipe_path
from cobra_db.filesystem import get_instance_path
from cobra_db.model import FileSource, ImageMetadata
from cobra_db.mongo_dao import Connector, ImageMetadataDao
from cobra_db.scripts.utils import add_args_to_iterable, batcher


def parse_arguments(raw_args: List[str]) -> str:
    """Parse the arguments for the pseudonymization process

    :param raw_args: arguments from sys.argv[1:]
    :return: a namespace with the required cfg
    """
    parser = ArgumentParser(
        "Pseudonymize ImageMetadata",
        description="""
Given an ImageMetadata collection with real data, create a second database with the
pseudonymized version. The files will be pseudonymized and stored according to the
suggested filesystem paths. Additionally, docs that are correctly pseudonymized will
be updated with a `deid_file_source` field that allows to check where the data went.""",
    )

    parser.add_argument(
        "config_path",
        type=str,
        help="the config file to be used.\
        For an example of the config file, please read the documentation how-to guide",
    )

    args = parser.parse_args(raw_args)
    return Namespace(**parse_config(args.config_path, default_value=None))


def im_meta_generator(src_mongo: dict, query: dict = {}):
    """Generator of im_meta dicts

    :param src_mongo: kwargs for the Connector
    :param query: query to select subset of the ImageMetadata collection
    :yield: dict with image metadata.
    """

    connector = Connector(**src_mongo)
    im_dao = ImageMetadataDao(connector)
    try:
        logging.info("Starting generator of ImageMetadata")
        cursor = im_dao.collection.find(
            query,
            {
                "_id": True,
                "_metadata": True,
                "dicom_tags": {"Not used": "Not used"},
                "file_source": True,
                "study_id": True,
                "series_id": True,
                "aka_file_sources": True,
            },
        )
        for x in cursor:
            yield x
    finally:
        cursor.close()


def im_meta_from_virtual_dataset(
    ds: pydicom.Dataset, file_source: FileSource, _id: ObjectId = None
):
    """An alternative version of ImageMetadata.from_dataset that does not require a
    ds.filename

    :param ds: Dicom dataset
    :param file_source: Source of the dataset
    :param _id: the id that will be stored in the dst_mongo
    """
    json_dict = ds.to_json_dict(suppress_invalid_tags=True)
    dicom_tags = DatasetMod.tags_to_keywords(json_dict)
    return ImageMetadata(
        _id=_id,
        series_id=None,
        _metadata=None,
        file_source=file_source,
        dicom_tags=dicom_tags,
    )


def process_batch(args):
    im_metas, cfg = args

    src_im_dao = ImageMetadataDao(Connector(**cfg.src_mongo))
    dst_im_dao = ImageMetadataDao(Connector(**cfg.dst_mongo))
    try:
        processed = 0
        seen = 0
        for im_meta in im_metas:
            seen += 1
            try:
                processed += process_im_meta(im_meta, cfg, src_im_dao, dst_im_dao)
            except IndexError:
                pass
    finally:
        src_im_dao.connector.close()
        dst_im_dao.connector.close()

    return seen, processed


def process_im_meta(
    im_meta_dict: dict, cfg, src_im_dao: ImageMetadataDao, dst_im_dao: ImageMetadataDao
):
    im_meta = ImageMetadata.from_dict(im_meta_dict)
    # get source file
    src_filepath = im_meta.get_local_filepath(cfg.mount_paths)
    try:
        src_ds = pydicom.read_file(src_filepath)
    except Exception as e:
        logging.error(f"Could not read {im_meta._id} - {e}")
        return 0

    # pseudonymize
    try:
        deid_ds = cfg.deider.pseudonymize(src_ds)
    except ValueError as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0
    except KeyError as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0
    except BytesLengthException as e:
        logging.error(f"Could not pseudonymize {im_meta._id} - {e}")
        return 0

    # get path where it will be saved
    dst_rel_dir = cfg.dst_rel_dir
    if dst_rel_dir is None:
        dst_rel_dir = ""
    dst_rel_filepath = os.path.join(dst_rel_dir, get_instance_path(deid_ds))
    dst_file_source = FileSource(cfg.dst_drive_name, dst_rel_filepath)
    if dst_file_source.drive_name in [
        x.get("drive_name") for x in im_meta_dict.get("aka_file_sources", [])
    ]:
        logging.debug("This file has already been created.")
        raise IndexError("This file has already been created.")
    dst_filepath = dst_file_source.get_local_filepath(cfg.mount_paths)

    os.makedirs(os.path.dirname(dst_filepath), exist_ok=True)
    try:
        pydicom.write_file(dst_filepath, deid_ds)
    except ValueError as e:
        logging.error(f"{e}")

    # the _id of the dataset is stored with the same database _id so that a super user
    # with access to both collections, can quickly check that everything is okay.
    try:
        del deid_ds.PixelData
    except AttributeError as e:
        logging.error(f"{e}")

    deid_im_meta = im_meta_from_virtual_dataset(deid_ds, dst_file_source, im_meta._id)

    # Keep record of what happened
    try:
        dst_im_dao.insert_one(deid_im_meta)
    except DuplicateKeyError as e:
        logging.error(f"{e}")
        return 0
    src_im_dao.add_aka(im_meta._id, dst_file_source)
    return 1


def get_required_drive_names(src_mongo, query):
    connector = Connector(**src_mongo)
    im_dao = ImageMetadataDao(connector)
    logging.info("Searching images to obtain required drives for query")
    cursor = im_dao.collection.aggregate(
        [{"$match": query}, {"$sortByCount": "$file_source.drive_name"}]
    )
    drive_names = []
    images = 0
    for doc in cursor:
        drive_names.append(doc["_id"])
        images += doc["count"]
    return drive_names, images


def check_mount_paths(mount_paths, required_drive_names, dst_drive_name):
    missing = set(required_drive_names) - set(mount_paths.keys())
    assert missing == set(), f"Missing configuration for drive_names: {missing}"
    assert mount_paths.get(dst_drive_name) is not None, "Missing dst_drive_name"


def single_proc(batches, n_proc=1):
    for batch in batches:
        seen, processed = process_batch(batch)
        yield seen, processed


def multi_proc(batches, n_proc):
    pool = Pool(n_proc)
    for seen, processed in pool.imap_unordered(process_batch, batches):
        yield seen, processed


def query_mux(query, image_ids: List[ObjectId] = None):
    """Select one of query or image_ids"""
    if image_ids is not None:
        assert query == {}, "Cannot have a query and a list of images at the same time"
        query = {"_id": {"$in": image_ids}}
    return query


def recipe_mux(base: bool, mr: bool, recipe: Union[str, List[str]]):
    """Select the correct recipe according to the configurations"""
    recipes = []
    if base:
        print(f"Using VAIB recipe from: {base_recipe_path}")
        recipes.append(base_recipe_path)
    if mr:
        print(
            f"Using additional to VAIB recipe targeted to MR studies from: {mr_recipe_path}"
        )
        recipes.append(mr_recipe_path)
    if type(recipe) == str:
        recipes.append(recipe)
    if type(recipe) == list:
        recipes = recipes + recipe
    for r in recipes:
        assert os.path.exists(r), f"Deid recipe path does not exist: {r}"
    if len(recipes) == 0:
        return None
    logging.info(f"Using recipes: {recipes}")
    return recipes


def main(args=None, image_ids: List[ObjectId] = None):
    if args is None:
        # Allows testing main from pytest.
        args = sys.argv[1:]

    # read config file
    cfg = parse_arguments(args)
    cfg.query = query_mux(cfg.query, image_ids)

    # Check how many files will be processed and if the cfg is enough
    required_drive_names, total_imgs = get_required_drive_names(
        cfg.src_mongo, cfg.query
    )
    check_mount_paths(cfg.mount_paths, required_drive_names, cfg.dst_drive_name)
    im_meta_gen = im_meta_generator(cfg.src_mongo, cfg.query)
    logging.info("Saving list of files to be processed")
    imgs = list(im_meta_gen)  # save as list to avoid losing the
    batches = batcher(imgs, batch_size=cfg.batch_size)

    # Due to a weird configuration system of the deid library this has to be loaded
    # after setting the MESSAGELEVEL env var.
    os.environ["MESSAGELEVEL"] = cfg.logging_level

    cfg.deider = Deider(
        cfg.hash_secret,
        recipe_mux(
            cfg.deid_default_recipes["base"],
            cfg.deid_default_recipes["mr"],
            cfg.user_recipe_path,
        ),
        logging_level=cfg.logging_level,
    )

    batches = add_args_to_iterable(batches, cfg)
    pbar_seen = tqdm(total=total_imgs, desc="Images seen")
    pbar_processed = tqdm(total=total_imgs, desc="Images processed")
    if cfg.n_proc == 1:
        process_func = single_proc
    if cfg.n_proc > 1:
        process_func = multi_proc
    if cfg.n_proc < 1:
        raise ValueError("n_proc must be bigger than 0")
    for seen, processed in process_func(batches, cfg.n_proc):
        pbar_processed.update(processed)
        pbar_seen.update(seen)



import logging
import os
from copy import copy
from dataclasses import dataclass, fields, is_dataclass
from datetime import datetime, timezone
from typing import Callable, List, Tuple, Type, Union

import numpy as np
import pydicom
from bson import ObjectId
from pydicom.dataset import Dataset

from cobra_db import __version__
from cobra_db.dataset_mod import DatasetMod
from cobra_db.enums import Modality, PregnancyStatus
from cobra_db.utils import (
    intersect_dicts,
    parse_AS_as_int,
    parse_DA_as_datetime,
    parse_DA_TM_as_datetime,
)


class BaseObj:
    """Defines the interface to load the object from dict and how to export it to dict."""

    @staticmethod
    def expand(value):
        if isinstance(value, BaseObj):
            return value.to_dict()
        else:
            return value

    def _to_dict(self):
        if is_dataclass(self):
            return self.__dict__
        else:
            raise NotImplementedError

    def to_dict(self):
        """expand the nested objects and remove all keys with None value"""
        d = {k: self.expand(v) for k, v in self._to_dict().items() if v is not None}
        return d

    @classmethod
    def from_dict(cls, d: dict):
        raise NotImplementedError(
            f"This method should be implemented in the child {type(cls)}"
        )


class Embedded(BaseObj):
    """Base class for an embedded doc (that doesn't get its own _id)"""

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)


@dataclass
class Metadata(Embedded):
    """Stores information of when the db doc was created, modified, and the data model
    version.
    Should be used in any doc with _id that is stored in the database.
    """

    model_version: str
    created: datetime
    modified: datetime = None
    project_name: str = None

    @classmethod
    def create(cls):
        return cls(model_version=__version__, created=datetime.now(timezone.utc))


@dataclass
class Entity(BaseObj):
    """Base class for all entities that are stored in the db with an _id, i.e. that
    require a Dao.
    Usually parent of a ``@dataclass`` class
    """

    _id: ObjectId
    _metadata: Metadata

    def __repr__(self):
        header = f"<{self.__class__.__name__}>"
        body = copy(self.__dict__)
        return f"{header}\n{body}\n"

    def __post_init__(self):
        """Automatically add db_metadata if missing"""
        if self._metadata is None:
            self._metadata = Metadata.create()
        if isinstance(self._metadata, dict):
            self._metadata = Metadata(**self._metadata)

    @classmethod
    def from_dict(cls, obj_dict: dict):
        obj = cls(**obj_dict)
        for f in fields(obj):
            typ = f.type
            if type(typ) != type:
                # TODO: not handled yet!!!
                pass
            else:
                val_dict = obj_dict.get(f.name, None)
                if val_dict is not None and issubclass(typ, BaseObj):
                    obj.__setattr__(f.name, typ.from_dict(val_dict))
        return obj


class Source:
    """Base class for a pointer to more information."""


@dataclass
class DicomEntity(Entity):
    """Base class for all entities that contain dicom_tags."""

    dicom_tags: dict

    @staticmethod
    def optional(ds: Dataset, tag_name: str, parser: Callable = None, default=None):
        """
        Get tag value from dataset, if missing return default value.
        """
        value = ds.get(tag_name, None)
        if value is not None and value != "":
            if parser is not None:
                value = parser(value)
            return value
        else:
            return default

    @staticmethod
    def choose_first(ds, *tag_names, parser, default):
        """
        Choose first tag that does not return None,
        """
        for tag_name in tag_names:
            tag = DicomEntity.optional(ds, tag_name, parser, None)
            if tag is not None:
                return tag
        return default

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """Abstract method. How to create an instance from a pydicom Dataset"""
        raise NotImplementedError(f"Should be implemented by child {type(cls)}")

    def get_tag(self, keyword: str, default=None):
        tag = self.dicom_tags.get(keyword, None)
        if tag is not None:
            value = tag.get("Value", default)
            if not value == default:
                if isinstance(value, list):
                    if len(value) == 1:
                        return value[0]  # most of the cases are this
                return value
        else:
            return default

    def get_acquisition_datetime(self):
        time = self.get_tag("AcquisitionTime")
        time = "000000" if time is None else time
        date = self.get_tag("AcquisitionDate")
        date = "00000000" if date is None else date
        return parse_DA_TM_as_datetime(date, time)


@dataclass
class EntitySource(Embedded, Source):
    """Pointer to another Entity instance in the database.
    Should only be used when there's ambiguity about the referenced class/collection.
    """

    name: str
    description: str
    _entity_type: Union[Type[Entity], str]
    _id: ObjectId

    def __post_init__(self):
        if not type(self._entity_type) is str:
            self._entity_type = self._entity_type.__name__

    def to_dict(self):
        return self.__dict__


@dataclass
class FileSource(Embedded, Source):
    """Pointer to a file in the filesystem.
    rel_path is relative to the place where drive_name is mounted.
    This is because drive_name can be mounted in different paths according
    to the machine and we don't have another way of referencing a drive.
    filename is automatically stored for easy indexing and querying without having to
    manipulate the rel_path to extract the filename.
    """

    drive_name: str
    rel_path: str
    filename: str = None  # automatically set when rel_path is set in __post_init__

    def get_local_filepath(self, mount_paths: dict) -> str:
        """Returns the filepath according to mount_paths"""
        return os.path.join(mount_paths[self.drive_name], self.rel_path)

    def __post_init__(self):
        if self.filename is None:
            self.filename = os.path.basename(self.rel_path)

    @classmethod
    def from_mount_paths(cls, filepath: str, mount_paths: dict):
        """create an instance from the filepath and the mount_paths.

        :param filepath: _description_
        :param mount_paths: _description_
        """
        for drive_name, mount_path in mount_paths.items():
            rel_path = os.path.relpath(filepath, mount_path)
            # TODO: this method will raise ValueError in Windows
            if not rel_path.startswith(".."):  # file in the mount_path
                return cls(drive_name=drive_name, rel_path=rel_path)
        raise ValueError(f"{filepath} is not in any of the mount paths: {mount_paths}")


@dataclass
class Patient(Entity):
    """Represents a unique person that has gone through at least one study.
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.2.2.html
    https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.2.3.html
    """

    anon_id: str
    hidden: bool
    birth_date: datetime = None

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """
        Create an instance of the entity from a DICOM dataset
        """
        return cls(
            _id=None,
            _metadata=None,
            anon_id=ds.PatientID,
            birth_date=DicomEntity.optional(
                ds, "PatientBirthDate", parse_DA_as_datetime, None
            ),
            hidden=False,
        )

    def __eq__(self, other: "Patient") -> bool:
        return self.anon_id == other.anon_id


@dataclass
class RadiologicalStudy(DicomEntity):
    """Aggregation of multiple dicom instances by the combination of patient and date."""

    study_uid: List[str]  # (0020,0010) LO StudyInstanceUID
    date: datetime  # (0008,0020) DA StudyDate + (0008,0030) TM StudyTime
    series_count: int  # (0020,0011) IS SeriesNumber, Overriden after grouping because
    # the tags are not reliable.
    modality: List[Modality] = None  # (0008,0060) CS Modality
    # SOPClassUID, the meaning can be found in pydicom._uid_dict.UID_dictionaty
    sop_class: List[str] = None
    description: str = None  # (0008,1030) LO StudyDescription
    patient_id: ObjectId = None
    accession_number: List[
        str
    ] = None  # Used to link the study with other administrative registers.
    patient_age: int = None  # in years (0010,1010) AS PatientAge
    patient_weight: float = None  # (0010,1030) PatientWeight
    patient_height: float = None  # (0010,1020) PatientSize
    patient_pregnancy_status: PregnancyStatus = None  # (0010,21C0)

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """Create an instance of the entity from a DICOM dataset"""
        try:
            modality = [cls.optional(ds, "Modality", Modality, None)]
        except ValueError as e:
            logging.error(f"{e}")
            modality = None

        return cls(
            _id=None,
            _metadata=None,
            study_uid=cls.optional(ds, "StudyInstanceUID", default=None),
            modality=modality,
            sop_class=[cls.optional(ds, "SOPClassUID", str, None)],
            date=parse_DA_TM_as_datetime(
                cls.optional(ds, "StudyDate", str, "00000000"),
                cls.optional(ds, "StudyTime", str, "000000"),
            ),
            description=cls.optional(ds, "StudyDescription"),
            accession_number=cls.optional(ds, "AccessionNumber", str, None),
            patient_age=cls.optional(ds, "PatientAge", parse_AS_as_int, None),
            patient_weight=cls.optional(ds, "PatientWeight", float, None),
            patient_height=cls.optional(ds, "PatientSize", float, None),
            patient_pregnancy_status=(
                cls.optional(ds, "PatientPregnancyStatus", PregnancyStatus, None)
            ),
            series_count=cls.optional(ds, "SeriesNumber", int, None),
            dicom_tags=DatasetMod.tags_to_keywords(ds.to_json_dict()),
        )

    @classmethod
    def from_dict(cls, d: dict):
        """Convert a mongodb dict into a RadiologicalStudy object"""
        return cls(**d)

    def __eq__(self, other: "RadiologicalStudy") -> bool:
        return self.anon_study_id == other.anon_study_id


@dataclass
class RadiologicalSeries(DicomEntity):
    study_id: ObjectId
    date: datetime = None  # (0008,0021) DA SeriesDate + (0008,0031) TM SeriesTime
    series_uid: str = None  # (0020,000E) UI SeriesInstanceUID
    description: str = None  # 0008103E
    # Overrided by the count in the ImageMetadata collection
    image_count: int = None  # (0020,0013) IS InstanceNumber.
    image_shape: Tuple[int, int] = None  # If the dicom_tags of all images agree.
    protocol_name: str = None  # (0018,1030) LO ProtocolName,
    # other possible tags 00180018, 0008103E, 00181081, 00181250, 00180024, 00181030

    def __repr__(self):
        return super().__repr__()

    @classmethod
    def from_dataset(cls, ds: Dataset):
        """
        Create an instance of the entity from a DICOM dataset
        """
        return cls(
            _id=None,
            _metadata=None,
            series_uid=cls.optional(ds, "SeriesInstanceUID", str, None),
            date=parse_DA_TM_as_datetime(
                cls.optional(ds, "SeriesDate", str, "00000000"),
                cls.optional(ds, "SeriesTime", str, "000000"),
            ),
            study_id=None,
            image_count=cls.optional(ds, "InstanceNumber", int, None),
            protocol_name=cls.optional(ds, "ProtocolName", str, None),
            description=cls.optional(ds, "SeriesDescription", str, None),
            image_shape=(
                cls.optional(ds, "Rows", int, None),
                cls.optional(ds, "Columns", int, None),
            ),
            dicom_tags=DatasetMod.tags_to_keywords(ds.to_json_dict()),
        )

    @classmethod
    def from_images_metadata(cls, images_metadata: List["ImageMetadata"]):
        #  Shared properties are kept
        shared_dicom_tags = intersect_dicts(
            [im_meta.dicom_tags for im_meta in images_metadata]
        )
        _shared_dicom_dict = DatasetMod.keywords_to_tags(
            shared_dicom_tags
        )  # back to dicom hex tags
        shared_ds = Dataset.from_json(_shared_dicom_dict)
        return cls.from_dataset(shared_ds)

    def __eq__(self, other: "RadiologicalSeries") -> bool:
        return self.series_uid == other.series_uid


@dataclass
class ImageMetadata(DicomEntity):
    """Represents one DICOM image i.e. a DICOM instance with PixelData.
    Only the dicom headers are stored. While the pixel data is not stored.
    """

    dicom_tags: dict
    file_source: FileSource
    study_id: ObjectId = None
    series_id: ObjectId = (
        None  # is added by the method ImageMetadataDao.update_series_id()
    )
    aka_file_sources: List[FileSource] = None

    @classmethod
    def from_dataset(cls, ds: Dataset, mount_paths: dict):
        """Create one instance from a pydicom Dataset.

        :param ds: pydicom Dataset
        :param mount_paths: dict that maps the drive_names to the mount path of the
         drives that are in use.
        """
        # sometimes (<1%) the dataset contains mismatches on the VR. For example IS with
        # a float value.
        # in those cases the tag has to be fixed to be able to create the dict.
        json_dict = ds.to_json_dict(suppress_invalid_tags=True)
        dicom_tags = DatasetMod.tags_to_keywords(json_dict)
        return cls(
            _id=None,
            series_id=None,
            _metadata=None,
            file_source=FileSource.from_mount_paths(ds.filename, mount_paths),
            dicom_tags=dicom_tags,
        )

    def to_dataset(self) -> Dataset:
        """Creates a pydicom Dataset with only the metadata in dicom_tags.
        Useful to use pydicom interface when required.
        """
        dicom_tags = DatasetMod.keywords_to_tags(self.dicom_tags)
        return Dataset.from_json(dicom_tags)

    def get_checksum(self, mount_paths: dict) -> str:
        """Obtain the md5 hexdigest hash of the original file. This helps finding
        duplicated images.
        Warning: md5 is not cryptographycally secure.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: string hexdigest with the hash of the file.
        """
        import hashlib

        filepath = self.get_local_filepath(mount_paths)
        with open(filepath, "rb") as file:
            return hashlib.md5(file.read()).hexdigest()

    def get_local_filepath(self, mount_paths: dict) -> str:
        """Returns the filepath of the file_source according to mount_paths.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: string hexdigest with the hash of the file.
        """
        return self.file_source.get_local_filepath(mount_paths)

    def get_pixel_data(self, mount_paths) -> np.ndarray:
        """Load the file and extract the pixel_array from the pydicom dataset.

        :param mount_paths: a dictionary mapping the drive_name to the mount path in the
         local computer.
        :returns: the content of pydicom.Dataset.pixel_array
        """
        filepath = self.get_local_filepath(mount_paths)
        ds = pydicom.dcmread(filepath)
        return ds.pixel_array

    def get_file_size(self, mount_paths) -> int:
        """Get the size in bytes of the file"""
        filepath = self.get_local_filepath(mount_paths)
        return os.path.getsize(filepath)

    def __repr__(self):
        return super().__repr__()
